#!/usr/bin/python3
import argparse
import contextlib
import os
import subprocess
import tempfile
import urllib.request
from datetime import datetime


def read_file(path) -> str:
    """
    Return the content of the file on the given path.
    """
    with open(path) as f:
        return f.read()


def download_github_release_tarball(user: str, project: str, version: str) -> str:
    """
    Download a github release tarball of the given project.
    """
    project_url = f"https://github.com/{user}/{project}"
    tarball_url = f"{project_url}/archive/v{version}.tar.gz"

    local_path = f"{project}-{version}.tar.gz"
    urllib.request.urlretrieve(
        tarball_url,
        local_path
    )

    return local_path


@contextlib.contextmanager
def extracted_tarball(path: str):
    """
    Extract a tarball into a temporary directory.
    Yields the temporary directory.
    """
    with tempfile.TemporaryDirectory() as tempdir:
        subprocess.run(["tar", "-xf", path, "--directory", tempdir], check=True)
        yield tempdir


def merge_specfiles(upstream: str, downstream: str, version: str, author: str):
    """
    Merge the upstream specfile with the changelog from downstream and add a new
    changelog entry.
    """
    upstream_spec_lines = upstream.splitlines()
    downstream_spec_lines = downstream.splitlines()

    # Find where changelog starts in both specfiles
    changelog_start_in_up_spec = upstream_spec_lines.index("%changelog")
    changelog_start_in_down_spec = downstream_spec_lines.index("%changelog")

    # Create a new changelog entry
    date = datetime.now().strftime("%a %b %d %Y")
    changelog = f"""\
* {date} {author} - {version}-1
- New upstream release

"""

    # Join it all together:
    # Firstly, let's take upstream spec file including the %changelog directory
    # Then, put the newly created changelog entry
    # Finally, put there the changelog from the downstream spec file
    merged_lines = upstream_spec_lines[:changelog_start_in_up_spec + 1] + \
                   changelog.splitlines() + \
                   downstream_spec_lines[changelog_start_in_down_spec + 1:]

    return "\n".join(merged_lines) + "\n"


def update_distgit(user: str, project: str, version: str, author: str, pkgtool: str, release: str):
    """
    Update the dist-git for a new release.
    """
    specfile = f"{project}.spec"

    tarball = download_github_release_tarball(user, project, version)

    old_downstream_specfile = read_file(specfile)

    with extracted_tarball(tarball) as path:
        upstream_specfile = read_file(f"{path}/{project}-{version}/{specfile}")

    new_downstream_specfile = merge_specfiles(upstream_specfile, old_downstream_specfile, version, author)

    with open(specfile, "w") as f:
        f.write(new_downstream_specfile)

    release_arg = ["--release", release] if release else []

    subprocess.check_call([pkgtool, *release_arg, "new-sources", tarball])
    subprocess.check_call(["git", "add", ".gitignore", specfile, "sources"])

    commit_message = f"Update to {version}"
    subprocess.check_call(["git", "commit", "-m", commit_message])


if __name__ == "__main__":
    parser = argparse.ArgumentParser(allow_abbrev=False)
    parser.add_argument("--version", metavar="VERSION", type=str, help="version to be released to downstream",
                        required=True)
    parser.add_argument("--author", metavar="AUTHOR", type=str,
                        help="author of the downstream change (format: Name Surname <email@example.com>", required=True)
    parser.add_argument("--pkgtool", metavar="PKGTOOL", type=str, help="fedpkg, centpkg, or rhpkg", required=True)
    parser.add_argument("--release", metavar="RELEASE", type=str, help="distribution release (required only for centpkg)")
    args = parser.parse_args()

    if args.pkgtool not in ["fedpkg", "centpkg", "rhpkg"]:
        raise RuntimeError("--pkgtool must be fedpkg, centpkg, or rhpkg!")

    update_distgit(
        "osbuild",
        "osbuild-composer",
        args.version,
        args.author,
        args.pkgtool,
        args.release,
    )



"""entrypoint - Containerized OSBuild Composer

This provides the entrypoint for a containerized osbuild-composer image. It
spawns `osbuild-composer` on start and manages it until it exits. The main
purpose of this entrypoint is to prepare everything to be usable from within
a container.
"""

import argparse
import contextlib
import os
import socket
import subprocess
import sys


class Cli(contextlib.AbstractContextManager):
    """Command Line Interface"""

    def __init__(self, argv):
        self.args = None
        self._argv = argv
        self._exitstack = None
        self._parser = None

    def _parse_args(self):
        self._parser = argparse.ArgumentParser(
            add_help=True,
            allow_abbrev=False,
            argument_default=None,
            description="Containerized OSBuild Composer",
            prog="container/osbuild-composer",
        )

        # --[no-]composer-api
        self._parser.add_argument(
            "--composer-api",
            action="store_true",
            dest="composer_api",
            help="Enable the composer-API",
        )
        self._parser.add_argument(
            "--no-composer-api",
            action="store_false",
            dest="composer_api",
            help="Disable the composer-API",
        )
        self._parser.add_argument(
            "--composer-api-port",
            type=int,
            default=443,
            dest="composer_api_port",
            help="Port which the composer-API listens on",
        )

        # --[no-]local-worker-api
        self._parser.add_argument(
            "--local-worker-api",
            action="store_true",
            dest="local_worker_api",
            help="Enable the local-worker-API",
        )
        self._parser.add_argument(
            "--no-local-worker-api",
            action="store_false",
            dest="local_worker_api",
            help="Disable the local-worker-API",
        )

        # --[no-]remote-worker-api
        self._parser.add_argument(
            "--remote-worker-api",
            action="store_true",
            dest="remote_worker_api",
            help="Enable the remote-worker-API",
        )
        self._parser.add_argument(
            "--no-remote-worker-api",
            action="store_false",
            dest="remote_worker_api",
            help="Disable the remote-worker-API",
        )

        # --[no-]weldr-api
        self._parser.add_argument(
            "--weldr-api",
            action="store_true",
            dest="weldr_api",
            help="Enable the weldr-API",
        )
        self._parser.add_argument(
            "--no-weldr-api",
            action="store_false",
            dest="weldr_api",
            help="Disable the weldr-API",
        )

        # --[no-]dnf-json
        self._parser.add_argument(
            "--dnf-json",
            action="store_true",
            dest="dnf_json",
            help="Enable  dnf-json",
        )
        self._parser.add_argument(
            "--no-dnf-json",
            action="store_false",
            dest="dnf_json",
            help="Disable dnf-json",
        )
        self._parser.add_argument(
            "--dnf-json-port",
            type=int,
            default=0,
            dest="dnf_json_port",
            help="Specify the port dnf-json should listen on",
        )

        self._parser.set_defaults(
            builtin_worker=False,
            composer_api=False,
            local_worker_api=False,
            remote_worker_api=False,
            weldr_api=False,
            dnf_json=False,
        )

        return self._parser.parse_args(self._argv[1:])

    def __enter__(self):
        self._exitstack = contextlib.ExitStack()
        self.args = self._parse_args()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._exitstack.close()
        self._exitstack = None

    def _prepare_sockets(self):
        # Prepare all the API sockets that osbuild-composer expectes, and make
        # sure to pass them according to the systemd socket-activation API.
        #
        # Note that we rely on this being called early, so we get the correct
        # FD numbers assigned. We need FD-#3 onwards for compatibility with
        # socket activation (because python `subprocess.Popen` does not support
        # renumbering the sockets we pass down).

        index = 3
        sockets = []
        names = []

        # osbuild-composer.socket
        if self.args.weldr_api:
            print("Create weldr-api socket", file=sys.stderr)
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self._exitstack.enter_context(contextlib.closing(sock))
            sock.bind("/run/weldr/api.socket")
            sock.listen()
            sockets.append(sock)
            names.append("osbuild-composer.socket")

            assert(sock.fileno() == index)
            index += 1

        # osbuild-composer-api.socket
        if self.args.composer_api:
            print("Create composer-api socket on port {}".format(self.args.composer_api_port) , file=sys.stderr)
            sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            self._exitstack.enter_context(contextlib.closing(sock))
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
            sock.bind(("::", self.args.composer_api_port))
            sock.listen()
            sockets.append(sock)
            names.append("osbuild-composer-api.socket")

            assert(sock.fileno() == index)
            index += 1

        # osbuild-local-worker.socket
        if self.args.local_worker_api:
            print("Create local-worker-api socket", file=sys.stderr)
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self._exitstack.enter_context(contextlib.closing(sock))
            sock.bind("/run/osbuild-composer/job.socket")
            sock.listen()
            sockets.append(sock)
            names.append("osbuild-local-worker.socket")

            assert(sock.fileno() == index)
            index += 1

        # osbuild-remote-worker.socket
        if self.args.remote_worker_api:
            print("Create remote-worker-api socket", file=sys.stderr)
            sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            self._exitstack.enter_context(contextlib.closing(sock))
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
            sock.bind(("::", 8700))
            sock.listen(256)
            sockets.append(sock)
            names.append("osbuild-remote-worker.socket")

            assert(sock.fileno() == index)
            index += 1

        # Prepare FD environment for the child process.
        os.environ["LISTEN_FDS"] = str(len(sockets))
        os.environ["LISTEN_FDNAMES"] = ":".join(names)

        return sockets

    @staticmethod
    def _spawn_worker():
        cmd = [
            "/usr/libexec/osbuild-composer/osbuild-worker",
            "-unix",
            "/run/osbuild-composer/job.socket",
        ]

        env = os.environ.copy()
        env["CACHE_DIRECTORY"] = "/var/cache/osbuild-worker"
        env["STATE_DIRECTORY"] = "/var/lib/osbuild-worker"

        return subprocess.Popen(
            cmd,
            cwd="/",
            env=env,
            stdin=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
        )

    @staticmethod
    def _spawn_composer(sockets):
        cmd = [
            "/usr/libexec/osbuild-composer/osbuild-composer",
            "-verbose",
        ]

        # Prepare the environment for osbuild-composer. Note that we cannot use
        # the `env` parameter of `subprocess.Popen()`, because it conflicts
        # with the `preexec_fn=` parameter. Therefore, we have to modify the
        # caller's environment.
        os.environ["CACHE_DIRECTORY"] = "/var/cache/osbuild-composer"
        os.environ["STATE_DIRECTORY"] = "/var/lib/osbuild-composer"

        # We need to set `LISTEN_PID=` to the target PID. The only way python
        # allows us to do this is to hook into `preexec_fn=`, which is executed
        # by `subprocess.Popen()` after forking, but before executing the new
        # executable.
        preexec_setenv = lambda: os.putenv("LISTEN_PID", str(os.getpid()))

        return subprocess.Popen(
            cmd,
            cwd="/usr/libexec/osbuild-composer",
            stdin=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
            pass_fds=[sock.fileno() for sock in sockets],
            preexec_fn=preexec_setenv,
        )

    def _spawn_dnf_json(self):
        cmd = [
            "/usr/libexec/osbuild-composer/dnf-json",
        ]

        if self.args.dnf_json_port:
            sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            self._exitstack.enter_context(contextlib.closing(sock))
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
            sock.bind(("::", self.args.dnf_json_port))
            sock.listen()
        else:
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self._exitstack.enter_context(contextlib.closing(sock))
            os.makedirs("/run/osbuild-dnf-json/", exist_ok=True)
            sock.bind("/run/osbuild-dnf-json/api.sock")
            sock.listen()

        dnfenv = os.environ.copy()
        dnfenv["LISTEN_FDS"] = "1"
        dnfenv["LISTEN_FD"] = str(sock.fileno())

        return subprocess.Popen(
            cmd,
            cwd="/usr/libexec/osbuild-composer",
            stdin=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
            env=dnfenv,
            pass_fds=[sock.fileno()]
        )

    def run(self):
        """Program Runtime"""

        proc_composer = None
        proc_worker = None
        proc_dnf_json = None
        res = 0
        sockets = self._prepare_sockets()

        try:
            if self.args.builtin_worker:
                proc_worker = self._spawn_worker()

            if self.args.dnf_json:
                proc_dnf_json = self._spawn_dnf_json()

            if any([self.args.weldr_api, self.args.composer_api, self.args.local_worker_api, self.args.remote_worker_api]):
                proc_composer = self._spawn_composer(sockets)

            if proc_composer:
                res = proc_composer.wait()

            if proc_worker:
                if proc_composer:
                    proc_worker.terminate()
                proc_worker.wait()

            if proc_dnf_json:
                if proc_composer:
                    proc_dnf_json.terminate()
                proc_dnf_json.wait()

            return res
        except KeyboardInterrupt:
            if proc_worker:
                proc_worker.terminate()
                proc_worker.wait()
            if proc_dnf_json:
                proc_dnf_json.terminate()
                proc_dnf_json.wait()
            if proc_composer:
                proc_composer.terminate()
                res = proc_composer.wait()
        except:
            if proc_worker:
                proc_worker.kill()
            if proc_dnf_json:
                proc_dnf_json.kill()
            if proc_composer:
                proc_composer.kill()
            raise

        return res


if __name__ == "__main__":
    with Cli(sys.argv) as global_main:
        sys.exit(global_main.run())



#!/usr/bin/env python
# Trigger a webhook event for Schutzbot using AWS SQS.
import json
import os

import boto3
from botocore import UNSIGNED
from botocore.client import Config

WEBHOOK_PAYLOAD = os.environ.get("WEBHOOK_PAYLOAD")
EVENT_NAME = os.environ.get("EVENT_NAME")
SQS_QUEUE_URL = os.environ.get("SQS_QUEUE_URL")
SQS_REGION = os.environ.get("SQS_REGION")

sqs = boto3.client(
    'sqs',
    region_name=SQS_REGION,
    config=Config(
        signature_version=UNSIGNED
    )
)

payload = json.loads(WEBHOOK_PAYLOAD)
message = {
    'headers': {'X-Github-Event': EVENT_NAME},
    'payload': payload
}

response = sqs.send_message(
    QueueUrl=SQS_QUEUE_URL,
    MessageBody=json.dumps(message)
)



#!/usr/bin/python3
import json
import sys
import time

import requests


# Composer API for Koji uses a slightly different repository format
# that osbuild-composer does in /usr/share/osbuild-composer/repositories.
#
# This function does the conversion.
def composer_repository_to_koji_repository(repository):
    koji_repository = {
        "baseurl": repository["baseurl"]
    }

    if repository.get("check_gpg", False):
        koji_repository["gpgkey"] = repository["gpgkey"]

    return koji_repository


def compose_request(distro, koji, arch):
    with open(f"/usr/share/tests/osbuild-composer/repositories/{distro}.json") as f:
        test_repositories = json.load(f)

    repositories = [composer_repository_to_koji_repository(repo) for repo in test_repositories[arch]]
    image_requests = [{
            "architecture": "x86_64",
            "image_type": "qcow2",
            "repositories": repositories
        }]

    #TODO: Remove this condition once there is rhel9 support for AMI image type
    if distro != "rhel-90":
        image_requests.append({
            "architecture": "x86_64",
            "image_type": "ami",
            "repositories": repositories
        })

    req = {
        "name": "name",
        "version": "version",
        "release": "release",
        "distribution": distro,
        "koji": {
            "server": koji,
            "task_id": 1
        },
        "image_requests": image_requests
    }

    return req


def main(distro, arch):
    cr = compose_request(distro, "https://localhost:4343/kojihub", arch)
    print(json.dumps(cr))

    r = requests.post("https://localhost/api/composer-koji/v1/compose", json=cr,
                      cert=("/etc/osbuild-composer/worker-crt.pem", "/etc/osbuild-composer/worker-key.pem"),
                      verify="/etc/osbuild-composer/ca-crt.pem")
    if r.status_code != 201:
        print("Failed to create compose")
        print(r.text)
        sys.exit(1)

    print(r.text)
    compose_id = r.json()["id"]

    while True:
        r = requests.get(f"https://localhost/api/composer-koji/v1/compose/{compose_id}",
                         cert=("/etc/osbuild-composer/worker-crt.pem", "/etc/osbuild-composer/worker-key.pem"),
                         verify="/etc/osbuild-composer/ca-crt.pem")
        if r.status_code != 200:
            print("Failed to get compose status")
            print(r.text)
            sys.exit(1)
        status = r.json()["status"]
        print(status)
        if status == "success":
            print("Compose worked!")
            print(r.text)
            break
        elif status == "failure":
            print("compose failed!")
            print(r.text)
            sys.exit(1)
        elif status != "pending" and status != "running":
            print(f"unexpected status: {status}")
            print(r.text)
            sys.exit(1)

        time.sleep(10)

    r = requests.get(f"https://localhost/api/composer-koji/v1/compose/{compose_id}/logs",
                     cert=("/etc/osbuild-composer/worker-crt.pem", "/etc/osbuild-composer/worker-key.pem"),
                     verify="/etc/osbuild-composer/ca-crt.pem")
    logs = r.json()
    assert "image_logs" in logs
    assert type(logs["image_logs"]) == list
    assert len(logs["image_logs"]) == len(cr["image_requests"])


if __name__ == "__main__":
    if len(sys.argv) != 3:
        print(f"usage: {sys.argv[0]} DISTRO ARCH", file=sys.stderr)
        sys.exit(1)
    main(sys.argv[1], sys.argv[2])



#!/usr/bin/env python3
import argparse, subprocess

def launch_server(address, port, certdir):
  cmd = [
      "/usr/libexec/osbuild-composer/osbuild-mock-openid-provider",
      "-a", str.join(":", [address, port]),
      "-rsaPubPem", f"{certdir}/client-crt.pem",
      "-rsaPem", f"{certdir}/client-key.pem",
  ]
  print("Running oath server")
  return subprocess.run(
      cmd,
      cwd="/usr/libexec/osbuild-composer",
      stdin=subprocess.DEVNULL,
      stderr=subprocess.STDOUT,
  )

def main():
  parser = argparse.ArgumentParser()
  parser.add_argument("-a", "--address", help="IP address for the server", type=str, default="localhost")
  parser.add_argument("-p", "--port", help="Port for the server", type=str, default="8080")
  parser.add_argument("-c", "--certdir", help="The location dir of the certs", type=str, default="/etc/osbuild-composer")
  args = parser.parse_args()
  launch_server(args.address, args.port, args.certdir)

if __name__ == "__main__":
  main()
